
# This file was generated by 'versioneer.py' (0.23.dev0) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2022-07-01T13:21:48-0500",
 "dirty": false,
 "error": null,
 "full-revisionid": "ecd6b49ca988c6f6280a4372461cd11f4ad77c03",
 "version": "2.7.4"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
